//================================================================================
//
//  This code is the intellectual property of Jerry Jongerius (duckware.com)
//
//================================================================================

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.text.*;

public class AppPanel extends Panel {
    private AppFrame m_parent;
    private Object m_lock = new Lock();    //
    private boolean m_bRunning=true;       // app still running? (otherwise exit)
    private String m_bandinfo=null;        // wireless band info
    private String m_bandpower=null;       // dBm power level info
    private String m_routername=null;      // router name: eg: "R7800"
    private Athstats m_ath1;               // captured output from 'athstats' 1
    private Athstats m_ath2;               // captured output from 'athstats' 2 (for difference)
    private String m_sta=null;             // captured output from 'wlanconfig' (stations)
    private String m_arp=null;             // captured output from 'arp -a'
    private long m_ms=0;


    //----------------------------------------------------------------------
    public AppPanel( AppFrame parent ) {
        m_parent = parent;
        }
    //----------------------------------------------------------------------
    public void setRunning( boolean bRunning ) {
        m_bRunning = bRunning;
        G.notify(m_lock);
        }

    //----------------------------------------------------------------------
    public void go( String args[] ) throws Exception {
        try {
            go2(args);
            }
        catch (ConnectException e) {
            System.out.println( ""+e );
            System.out.println( "ERROR: Does router has 'Enable Telnet' checked in /debug.htm?" );
            }
        catch (SocketTimeoutException e) {
            System.out.println( ""+e );
            System.out.println( "ERROR: Connection to router lost." );
            }
        catch (NoRouteToHostException e) {
            System.out.println( ""+e );
            System.out.println( "ERROR: Is Ethernet/Wi-Fi down?" );
            }
        finally {
            m_parent.dispose();
            }
        }

    //----------------------------------------------------------------------
    private void go2( String args[] ) throws Exception {

        // process command line
        String i_host = null;
        String i_password = null;
        String i_num = null;
        for (int loop=0; loop<args.length; ++loop) {
            String s = args[loop];
            if (s.equals("-debug")) {
                G.g_debug = true;
                }
            else if (null==i_host) {
                i_host = s;
                InetAddress.getByName(i_host);  // verify valid host/ip
                }
            else if (null==i_password) {
                i_password = s;
                }
            else if (null==i_num) {
                i_num = ""+parseNum(s);
                }
            else  {
                throw new RuntimeException("Unexpected command line argument '"+s+"'" );
                }
            }

        // on good command line (have host/password), continue
        if ((null!=i_host) && (null!=i_password)) {
            String i_wifi = "wifi" + ((null!=i_num)?i_num:"0");
            String i_ath =  "ath"  + ((null!=i_num)?i_num:"0");

            RouterTelnet rti = new RouterTelnet(i_host, i_password);
            rti.connect();

            m_parent.show();

            // get wireless band info (and power level in that band)
            if (true) {
                StringTokenizer tok = new StringTokenizer(rti.runCommand("iwconfig"),"\r\n");
                String findband = i_ath+" ";
                while (tok.hasMoreTokens()) {
                    String line = tok.nextToken();
                    if (line.startsWith(findband)) {
                        m_bandinfo = line.substring(findband.length()).trim();
                        break;
                        }
                    }
                // power for band is immediately 'next'
                String findpower = "Tx-Power:";
                while (tok.hasMoreTokens()) {
                    String line = tok.nextToken();
                    int nPos = line.indexOf(findpower);
                    if (nPos>0) {
                        m_bandpower = line.substring(nPos+findpower.length());
                        break;
                        }
                    }
                }

            // get router hostname
            if (true) {
                StringTokenizer tok = new StringTokenizer(rti.runCommand("hostname"),"\r\n");
                if (1==tok.countTokens()) {
                    String name = tok.nextToken();
                    if ((name.length()>3) && (name.length()<15)) {
                        m_routername = name;
                        }
                    }
                }

            // get 'MCS STATS'/etc over and over
            while (m_bRunning) {
                long t1 = G.tick();
                String ath = rti.runCommand("athstats -i "+i_wifi);
                String sta = rti.runCommand("wlanconfig "+i_ath+" list sta");  // can be blank with no stations
                String arp = rti.runCommand("arp -a");
                sta = G.replace(sta, "23069M", "   ...");
                sta = G.replace(sta, "IEEE80211_MODE_", "");
                Athstats newath2 = parseAthStats(ath);

                m_ath1 = m_ath2;
                m_ath2 = newath2;
                m_sta = sta;
                m_arp = arp;
                m_ms = G.tick()-t1;
                repaint();
                G.wait(m_lock,1500);  // if too quick, athstats numbers don't change (diff all zeros)
                }
            }

        // otherwise, display usage information
        else {
            System.out.println( "MCS Spy - by Jerry Jongerius (duckware.com)" );
            System.out.println( "" );
            System.out.println( "Usage: java -jar mcsspy.jar {-debug} <host> <password> {interface}" );
            System.out.println( "" );
            System.out.println( "  -debug    - turns on detailed debug output [optional]" );
            System.out.println( "  host      - ip address of router" );
            System.out.println( "  password  - password to router web interface" );
            System.out.println( "  interface - 0|1{|2} - 0=5GHz, 1=2.4GHz, etc [optional]" );
            System.out.println( "" );
            System.out.println( "Pulls and graphs 'MCS STATS' in real-time from (Qualcomm Atheros) router.  MCS" );
            System.out.println( "bar graphs are packets/sec (blue=recent; gray=historical).  See also" );
            System.out.println( "duckware.com/wifi and duckware.com/phytables" );
            System.out.println( "" );
            System.out.println( "NOTE: REQUIRES router's telnet interface enabled" );
            System.out.println( " - Netgear: Check 'Enable Telnet' in http://<router>/debug.htm" );
            }
        }

    //======================================================================
    //======================================================================

    //----------------------------------------------------------------------
    // parse output from "athstats" for MCS statistics
    private Athstats parseAthStats( String s ) throws IOException {
        Athstats ret = new Athstats();
        DataInputStream dis = new DataInputStream(new ByteArrayInputStream(G.s2b(s)));
        String line;
        while (null!=(line=dis.readLine())) {
            if (line.indexOf("Rx MCS STATS:")>0) {
                ret.rxmcs = tuple(skipPastColon(dis.readLine()) + skipPastColon(dis.readLine()));
                G.dout( "Rx:"+t2s(ret.rxmcs) );
                }
            else if (line.indexOf("Tx MCS STATS:")>0) {
                ret.txmcs = tuple(skipPastColon(dis.readLine()) + skipPastColon(dis.readLine()));
                G.dout( "Tx:"+t2s(ret.txmcs) );
                }
            else if (line.startsWith("noise floor")) {
                ret.nNoiseFloor = parseNum(skipPastColon(line));
                }
            else if (line.startsWith("ast_rx_bytes")) {
                ret.rxbytes = parseNum(skipPastColon(line));
                }
            else if (line.startsWith("ast_tx_bytes")) {
                ret.txbytes = parseNum(skipPastColon(line));
                }
            }
        return ret;
        }
    //----------------------------------------------------------------------
    private Dimension paintMCS( Graphics g, Color cBar, int x, int y, String label, int mcs[], int back[], int scale, String bps ) {
        int _y = y;

        // header
        g.setFont( new Font("Arial", Font.BOLD, 16) );
        y += g.getFontMetrics().getAscent();
        g.setColor( Color.BLUE );
        g.drawString( label, x, y );
        y += g.getFontMetrics().getAscent()+3;

        // prep for rest of text
        g.setFont( new Font("Arial", Font.BOLD, 16) );
        FontMetrics fm = g.getFontMetrics();
        int h = fm.getAscent();

        int backscale = Math.max(100,asum(back));
        g.setColor( Color.BLACK );
        int bb = fm.stringWidth("0: ");
        for (int loop=0; loop<mcs.length; ++loop) {

            g.setColor( new Color(200,200,200) );
            g.fillRect( x+bb, y-h, back[loop]*100/backscale, h );

            g.setColor( cBar );
            int per = mcs[loop]*100/scale;
            g.fillRect( x+bb, y-h, Math.max(mcs[loop]>1?1:0,per), h );

            g.setColor( (mcs[loop]>0) ? Color.BLUE : Color.GRAY );
            g.drawString( loop+":", x, y );
            if (mcs[loop]>0) {
                g.drawString( ""+mcs[loop], x+bb+per+10, y );
                }
            y += h+8;
            }

        g.setColor( Color.GRAY );
        g.drawString( bps, x+bb, y );
        y += h+8;

        return new Dimension( bb+10+100+fm.stringWidth("00000"), y-_y );
        }
    //----------------------------------------------------------------------
    // want to prefix each station line with ip address of that station (from arp info)
    // assumes: station starts with mac; arp lines start with IP, and contains mac in line
    private String arpfor( String line, int loop ) {
        String ret = loop>0 ? "" : ".IP";
        if ((null!=m_arp) && (loop>0)) {
            String mac = line.substring(0,line.indexOf(' ')+1).trim();
            int l1 = mac.length()>0 ? m_arp.indexOf(mac) : -1;
            int l2 = l1>0 ? m_arp.lastIndexOf('\n',l1) : -1;
            String ip = (l1>0 && l2>0) ? m_arp.substring(l2+1,m_arp.indexOf(' ',l2+1)+1).trim() : "";
            int dot = ip.lastIndexOf('.');
            ret += ((dot>0)?ip.substring(dot):" -");
            }
        return (ret+"     ").substring(0,5);
        }
    //----------------------------------------------------------------------
    private int paintStations( Graphics g, int x, int y ) {
        int _y = y;
        StringTokenizer tok = new StringTokenizer((null!=m_sta)?m_sta:"", "\r\n");
        if (tok.countTokens()>0) {
            g.setFont( new Font("Arial", Font.BOLD, 16) );
            int h = g.getFontMetrics().getAscent();
            y += h;
            String line;
            g.setColor( Color.BLUE );
            String header="STATION Information";
            g.drawString( header, x, y );
            g.setColor( Color.BLACK );
            g.setFont( new Font("Courier", Font.PLAIN, 14) );
            h = g.getFontMetrics().getAscent();
            y += h+8;
            for (int loop=0; tok.hasMoreTokens(); ++loop) {
                line = tok.nextToken();
                g.drawString( arpfor(line,loop)+line, x, y );
                y += h+8;
                }
            }

        return y-_y;
        }

    //----------------------------------------------------------------------
    // OVERRIDE
    public void update( Graphics g ) {
        paint( g );
        }
    //----------------------------------------------------------------------
    // OVERRIDE
    public void paint( Graphics g ) {
        if (DoubleBuffer.go(this,g)) {
            paint2( g );
            }
        }

    //----------------------------------------------------------------------
    public void paint2( Graphics g ) {
        g.setFont( new Font("Arial", Font.BOLD, 16) );
        int xx = 10;
        int yy = 10;

        if (true) {
            g.setFont( new Font("Arial", Font.BOLD, 22) );
            String routername = (null!=m_routername) ? m_routername : "Router";
            String bandinfo = (null!=m_bandinfo) ? "  --  "+m_bandinfo : "";
            String noisefloor = ((null!=m_ath2) && (m_ath2.nNoiseFloor<0)) ? "  --  Noise Floor "+m_ath2.nNoiseFloor : "";
            String powerlevel = (null!=m_bandpower)?" -- "+m_bandpower:"";
            String title = routername + bandinfo + noisefloor + powerlevel;
            FontMetrics fm = g.getFontMetrics();
            yy += fm.getAscent();
            Dimension d = size();
            g.setColor( Color.BLUE );
            g.drawString(title, xx, yy );
            yy += 8;
            }

        if ((null!=m_ath1) && (null!=m_ath2)) {
            long delta = m_ath2.ts-m_ath1.ts;
            int tx[] = adiff(m_ath1.txmcs, m_ath2.txmcs, delta);
            int rx[] = adiff(m_ath1.rxmcs, m_ath2.rxmcs, delta);
            if ((null!=tx) && (null!=tx)) {
                int scale = Math.max(100,Math.max(asum(tx),asum(rx)));
                String bps=tobps(m_ath2.txbytes-m_ath1.txbytes,m_ath2.ts-m_ath1.ts);
                Dimension d1 = paintMCS( g, Color.BLUE, xx, yy, "Tx MCS", tx, m_ath2.txmcs, scale, bps );
                bps = tobps(m_ath2.rxbytes-m_ath1.rxbytes,m_ath2.ts-m_ath1.ts);
                Dimension d2 = paintMCS( g, Color.BLUE, xx+d1.width+30, yy, "Rx MCS", rx, m_ath2.rxmcs, scale, bps );
                yy += Math.max(d1.height, d2.height);
                }
            }
        else {
            yy += 30;
            }

        yy += paintStations( g, xx, yy );

        //
        g.setFont( new Font("Arial", Font.PLAIN, 12) );
        g.setColor( Color.BLUE );
        g.drawString( "Gray bars are historical MCS usage since router reboot.  Blue bars are MCS usage over last moment. See duckware.com/mcs for MCS index to PHY speed tables.  ("+m_ms+" ms)", xx, size().height-10 );
        }

    //======================================================================
    //======================================================================
    //======================================================================


    //----------------------------------------------------------------------
    // want everything after first colon
    private static String skipPastColon( String s ) {
        int n = s.indexOf(":");
        return (n>0) ? s.substring(n+1).trim() : s;
        }
    //----------------------------------------------------------------------
    // convert string to tuple
    private static int[] tuple( String s ) {
        StringTokenizer tok = new StringTokenizer(s," ,");
        int num = tok.countTokens();
        int ret[] = new int[num];
        for (int loop=0; loop<num; ++loop) {
            ret[loop] = parseNum(tok.nextToken());
            }
        return ret;
        }
    //----------------------------------------------------------------------
    // sum all array entries
    private static int asum( int n[] ) {
        int sum=0;
        if (null!=n) {
            for (int loop=0; loop<n.length; ++loop) {
                sum += n[loop];
                }
            }
        return sum;
        }
    //----------------------------------------------------------------------
    // debug: tuple to String
    private static String t2s( int mcs[] ) {
      String ret="";
      for (int loop=0; loop<mcs.length; ++loop) {
          ret += ((loop>0)?",":"") + mcs[loop];
          }
      return ret;
      }
    //----------------------------------------------------------------------
    // array difference (with metrics prorated to 1 second)
    private static int[] adiff( int n1[], int n2[], long ms ) {
        int ret[] = null; 
        if ((null!=n1) && (null!=n2) && (ms>0)) {
            ret = new int[Math.min(n1.length,n2.length)];
            for (int loop=0; loop<ret.length; ++loop) {
                ret[loop] = (int)((n2[loop]-n1[loop])*1000L/ms);
                }
            }
        return ret;
        }
    //----------------------------------------------------------------------
    private static String tobps( int nBytes, long t ) {
        //long v = (long)(nBytes*8/(t/1000.0)/100000);
        //return (v/10)+"."+(v%10)+" Mbps";
        long v = (long)(nBytes*8/(t/1000.0));
        return NumberFormat.getNumberInstance(Locale.US).format(v)+" bps";
        }
    //----------------------------------------------------------------------
    // parse Atheros u_int32_t number -- put into Java int
    private static int parseNum( String s ) {
        return (int)Long.parseLong(s);
        }
    }
