//================================================================================
//
//  This code is the intellectual property of Jerry Jongerius (duckware.com)
//
//================================================================================

//
//  Connect to the router's telnet port and issue commands and pass back command
//  output.  We ignore telnet's do/don't/will/won't negotiation, because it is
//  not important and does not affect the output of the commands we will be
//  running.
//

import java.io.*;
import java.net.*;
import java.util.*;

public class RouterTelnet {
    private final static String NEWPROMPT="XXYYZZ: ";
    private String m_host;                 // host/ip of router
    private String m_password;             // password to router web interface
    private OutputStream m_os;             // telnet output stream
    private InputStream  m_is;             // telnet input stream

    //----------------------------------------------------------------------
    public RouterTelnet( String host, String password ) {
        m_host = host;
        m_password = password;
        }
    //----------------------------------------------------------------------
    // connect to remote router telnet
    public void connect() throws IOException {
        int nPort = 23;
        System.out.println( "Connecting to router "+m_host+" telnet port "+nPort );
        Socket s = new Socket( m_host, nPort );
        m_os = s.getOutputStream();
        m_is = s.getInputStream();
        // sign into router telnet (and change prompt)
        System.out.println( "Waiting for telnet password prompt..." );
        readUntilFound( "telnet password:" );
        System.out.println( "Entering password..." );
        sendLine( m_password );
        sendLine( "PS1='"+NEWPROMPT+"'" );
        readUntilPrompt();
        System.out.println( "Running..." );
        }

    //----------------------------------------------------------------------
    // run remote command and return output command generated
    public String runCommand( String cmd ) throws IOException {
        sendLine( cmd );
        readUntilFound( cmd+"\r\n" );    // discard remote echo
        G.dout( "ECHOCAN" );
        return readUntilPrompt();        // output until prompt
        }

    //======================================================================
    //======================================================================

    //----------------------------------------------------------------------
    private char readNextInputChar() throws IOException {
        int c = m_is.read();
        if (-1==c) {
            throw new IOException( "EOF" );
            }
        if (G.g_debug) {
            System.out.print( (char)c );
            //System.out.flush();
            }
        return (char)c;
        }
    //----------------------------------------------------------------------
    // read all output from router and stop when specific text is found
    private String readUntilFound( String find ) throws IOException {
        String s = "";
        while (!s.endsWith(find)) {
            s += readNextInputChar();
            }
        return s;
        }
    //----------------------------------------------------------------------
    // collect output from router; stop at (and discard) command prompt
    private String readUntilPrompt() throws IOException {
        G.dout( "+PSEARCH" );
        String s = "";
        while (!s.endsWith(NEWPROMPT) || (0!=m_is.available())) {
            s += readNextInputChar();
            }
        s = s.substring(0, Math.max(0,s.lastIndexOf("\r\n")));
        G.dout( "-PSEARCH" );
        return s;
        }
    //----------------------------------------------------------------------
    private void sendLine( String line ) throws IOException {
        G.dout( "SEND:"+line );
        m_os.write( G.s2b(line+"\r") );
        }
    }

